from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class BorrowerFieldAPIDTO(BaseModel):
    id: str = Field(alias="id")
    borrower_id: str = Field(alias="borrowerId")
    key: str = Field(alias="key")
    label: str = Field(alias="label")
    value: Any = Field(alias="value")
    data_type: str = Field(alias="dataType")
    history: List[Dict] = Field(alias="history")
    tags: List[str] = Field(alias="tags", default=[])
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateBorrowerFieldDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    form_id: Optional[str] = Field(alias="formId", default=None)
    reference_id: Optional[str] = Field(alias="referenceId", default=None)
    key: str = Field(alias="key")
    value: str = Field(alias="value")
    data_type: str = Field(alias="dataType")
    tags: List[str] = Field(alias="tags", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateBorrowerFieldDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    form_id: Optional[str] = Field(alias="formId", default=None)
    reference_id: Optional[str] = Field(alias="referenceId", default=None)
    value: Optional[str] = Field(alias="value")
    data_type: Optional[str] = Field(alias="dataType")
    tags: List[str] = Field(alias="tags", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BorrowerFieldSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "borrower-fields", header_builder, renew_token, BorrowerFieldAPIDTO.parse_obj(data))


class BorrowerFieldAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "borrower-fields", header_builder, renew_token, BorrowerFieldAPIDTO.parse_obj(data))


class BorrowerFieldsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=BorrowerFieldSync,
                         retrieve_data_model=BorrowerFieldAPIDTO,
                         create_data_model=CreateBorrowerFieldDTO,
                         update_data_model=UpdateBorrowerFieldDTO,
                         resource="borrower-fields")


class BorrowerFieldsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=BorrowerFieldAsync,
                         retrieve_data_model=BorrowerFieldAPIDTO,
                         create_data_model=CreateBorrowerFieldDTO,
                         update_data_model=UpdateBorrowerFieldDTO,
                         resource="borrower-fields")



from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class FormAPIDTO(BaseModel):
    id: str = Field(alias="id")
    tenant: str = Field(alias="tenant")
    form_id: str = Field(alias="formId")
    template_id: str = Field(alias="templateId")
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    is_finished: bool = Field(alias="isFinished", default=False)
    is_notified: bool = Field(alias="isNotified", default=False)
    last_ping_at: Optional[str] = Field(alias="lastPingAt", default=None)
    finished_at: Optional[str] = Field(alias="finishedAt", default=None)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class StartFormRequest(BaseModel):
    template_slug: str = Field(alias="templateSlug")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BorrowerSignUpRequest(BaseModel):
    tenant: str = Field(alias="tenant")
    persona: str = Field(alias="persona")
    form_id: str = Field(alias="formId")
    is_primary: bool = Field(alias="isPrimary", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BorrowerSignUpResponse(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    form_token: str = Field(alias="formToken")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class IdentityLookupResponse(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    points_of_contact: List[Dict] = Field(alias="pointsOfContact")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class FormSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "forms", header_builder, renew_token, FormAPIDTO.parse_obj(data))


class FormAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "forms", header_builder, renew_token, FormAPIDTO.parse_obj(data))


class FormsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, sync_resource=FormSync, retrieve_data_model=FormAPIDTO,
                         create_data_model=StartFormRequest, update_data_model=None, resource="forms")

    @retry_on_401
    def command_borrower_sign_up(self, borrower_sign_up_request: dict):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                f"/v1/{self.resource}/commands/borrower-sign-up",
                json=BorrowerSignUpRequest.parse_obj(borrower_sign_up_request).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            return BorrowerSignUpResponse(
                borrower_id=response.json().get("borrowerId"),
                form_token=response.headers.get("Authorization").split(" ")[1]
            )

    @retry_on_401
    def query_identity_lookup(self, tenant: str, key: str, value: str, form_id: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/{self.resource}/queries/identity-lookup",
                params={
                    "key": key,
                    "value": value,
                    "form-id": form_id,
                    "tenant": tenant
                },
                timeout=120
            )
            raise_for_status_improved(response)
            return IdentityLookupResponse.parse_obj(response.json())

    @retry_on_401
    def query_entity_value(self, borrower_id: str, entity_type: str, key: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/{self.resource}/queries/entity-value",
                params={
                    "borrower-id": borrower_id,
                    "key": key,
                    "entity-type": entity_type,
                },
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)
            return response.json().get("value")


class FormsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, async_resource=FormSync, retrieve_data_model=FormAPIDTO,
                         create_data_model=StartFormRequest, update_data_model=None, resource="forms")

    @retry_on_401_async
    async def command_borrower_sign_up(self, borrower_sign_up_request: dict):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                f"/v1/{self.resource}/commands/borrower-sign-up",
                json=BorrowerSignUpRequest.parse_obj(borrower_sign_up_request).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            return BorrowerSignUpResponse(
                borrower_id=response.json().get("borrowerId"),
                form_token=response.headers.get("Authorization").split(" ")[1]
            )

    @retry_on_401_async
    async def query_identity_lookup(self, tenant: str, key: str, value: str, form_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/{self.resource}/queries/identity-lookup",
                params={
                    "key": key,
                    "value": value,
                    "form-id": form_id,
                    "tenant": tenant
                },
                timeout=120
            )
            raise_for_status_improved(response)
            return IdentityLookupResponse.parse_obj(response.json())

    @retry_on_401_async
    async def query_entity_value(self, borrower_id: str, entity_type: str, key: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/{self.resource}/queries/entity-value",
                params={
                    "borrower-id": borrower_id,
                    "key": key,
                    "entity-type": entity_type,
                },
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)
            return response.json().get("value")



from pydantic import BaseModel, Field
from typing import Optional, List, Dict
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.cms.model.generics import GenericSyncModule, GenericAsyncModule
from altscore.cms.helpers import build_headers


class PartnerAPIDTO(BaseModel):
    id: str = Field(alias="partnerId")
    avatar: Optional[str] = Field(alias="avatar", default="")
    name: str = Field(alias="name")
    short_name: str = Field(alias="shortName")
    partner_id: str = Field(alias="partnerId")
    status: str = Field(alias="status")
    is_aggregator: bool = Field(alias="isAggregator")
    email: str = Field(alias="email")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class CreatePartnerDTO(BaseModel):
    name: str = Field(alias="name")
    short_name: str = Field(alias="shortName")
    email: str = Field(alias="email")
    tax_id: str = Field(alias="taxId")
    is_aggregator: Optional[bool] = Field(alias="isAggregator", default=False)
    avatar: Optional[str] = Field(alias="avatar", default="")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DPASettingsDefaults(BaseModel):
    amortization_type: Optional[str] = Field(alias="amortizationType", default=None)
    currency: Optional[str] = Field(alias="currency", default=None)
    flowExpirationMinutes: Optional[int] = Field(alias="flowExpirationMinutes", default=None)
    installments: Optional[int] = Field(alias="installments", default=None)
    repayEvery: Optional[int] = Field(alias="repayEvery", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Rate(BaseModel):
    period: Optional[int] = Field(alias="period", default=None)
    rate: Optional[str] = Field(alias="rate", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DPASettingssInterestRates(BaseModel):
    period: Optional[int] = Field(alias="period", default=None)
    rate: Optional[Rate] = Field(alias="rate", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DPASettingsPenalty(BaseModel):
    charge_code: Optional[str] = Field(alias="chargeCode", default=None)
    compute_every: Optional[int] = Field(alias="computeEvery", default=None)
    enabled: Optional[bool] = Field(alias="enabled", default=None)
    grace_period: Optional[int] = Field(alias="gracePeriod", default=None)
    rate: Optional[Rate] = Field(alias="rate", default=None)
    times_to_compute: Optional[int] = Field(alias="timesToCompute", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DPASettingsAPIDTO(BaseModel):
    defaults: Optional[DPASettingsDefaults] = Field(alias="defaults", default=None)
    disbursement: Optional[str] = Field(alias="disbursement", default=None)
    interest_rates: List[DPASettingssInterestRates] = Field(alias="interestRates", default=None)
    invoice_over_limit: Optional[float] = Field(alias="invoiceOverLimit", default=None)
    on_approve_flow_reserve_all_assigned_amount: Optional[bool] = \
        Field(alias="onApproveFlowReserveAllAssignedAmount", default=None)
    pentalties: List[DPASettingsPenalty] = Field(alias="pentalties", default=[])
    reserve_on_start: Optional[bool] = Field(alias="reserveOnStart", default=None)
    tax_rate: Optional[float] = Field(alias="taxRate", default=None)
    timezone: Optional[str] = Field(alias="timezone", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PartnerBase:

    def __init__(self, base_url):
        self.base_url = base_url


class PartnerAsync(PartnerBase):
    data: PartnerAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: PartnerAPIDTO):
        super().__init__(base_url)
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__})"


class PartnerSync(PartnerBase):
    data: PartnerAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: PartnerAPIDTO):
        super().__init__(base_url)
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.partner_id})"


class PartnersAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            async_resource=PartnerAsync,
            retrieve_data_model=PartnerAPIDTO,
            create_data_model=CreatePartnerDTO,
            update_data_model=None,
            resource="partners",
            resource_version="v2"
        )

    @retry_on_401_async
    async def me(self) -> PartnerAsync:
        async with httpx.AsyncClient(base_url=self.altscore_client._cms_base_url) as client:
            response = await client.get(
                "/v2/partners/me",
                # This is important to avoid infinite recursion
                headers=build_headers(self, partner_id=None),
                timeout=30
            )
            raise_for_status_improved(response)
            return PartnerAsync(
                base_url=self.altscore_client._cms_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=PartnerAPIDTO.parse_obj(response.json())
            )

    @retry_on_401_async
    async def update_dpa_settings(self, partner_id: str, settings: dict) -> DPASettingsAPIDTO:
        async with httpx.AsyncClient(base_url=self.altscore_client._cms_base_url) as client:
            settings = DPASettingsAPIDTO.parse_obj(settings)
            response = await client.patch(
                f"/v2/partners/{partner_id}/settings/dpa",
                json=settings.dict(by_alias=True, exclude_none=True),
                headers=build_headers(self),
                timeout=30
            )
            raise_for_status_improved(response)
            return DPASettingsAPIDTO.parse_obj(response.json())


class PartnersSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            sync_resource=PartnerSync,
            retrieve_data_model=PartnerAPIDTO,
            create_data_model=CreatePartnerDTO,
            update_data_model=None,
            resource="partners",
            resource_version="v2"
        )

    @retry_on_401
    def me(self) -> PartnerSync:
        with httpx.Client(base_url=self.altscore_client._cms_base_url) as client:
            response = client.get(
                "/v2/partners/me",
                # This is important to avoid infinite recursion
                headers=build_headers(self, partner_id="init"),
                timeout=30
            )
            raise_for_status_improved(response)
            return PartnerSync(
                base_url=self.altscore_client._cms_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=PartnerAPIDTO.parse_obj(response.json())
            )

    @retry_on_401
    def update_dpa_settings(self, partner_id: str, settings: dict) -> DPASettingsAPIDTO:
        settings = DPASettingsAPIDTO.parse_obj(settings)
        with httpx.Client(base_url=self.altscore_client._cms_base_url) as client:
            response = client.patch(
                f"/v2/partners/{partner_id}/settings/dpa",
                json=settings.dict(by_alias=True, exclude_none=True),
                headers=build_headers(self),
                timeout=30
            )
            raise_for_status_improved(response)
            return DPASettingsAPIDTO.parse_obj(response.json())



from pydantic import BaseModel, Field
from typing import Optional, Dict, List
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class AuthorizationAPIDTO(BaseModel):
    id: str = Field(alias="id")
    tenant: str = Field(alias="tenant")
    form_id: Optional[str] = Field(alias="formId")
    key: str = Field(alias="key")
    label: str = Field(alias="label")
    identity_key: str = Field(alias="identityKey")
    identity_value: str = Field(alias="identityValue")
    borrower_id: Optional[str] = Field(alias="borrowerId")
    expires_at: Optional[str] = Field(alias="expiresAt")
    ip_address: Optional[str] = Field(alias="ipAddress")
    policy_link: Optional[str] = Field(alias="policyLink")
    external_id: Optional[str] = Field(alias="externalId", default=None)
    tags: List[str] = Field(alias="tags", default=[])
    authorized_at: Optional[str] = Field(alias="authorizedAt")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")
    has_attachments: bool = Field(alias="hasAttachments", default=False)
    has_signatures: bool = Field(alias="hasSignatures", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True



class CreateAuthorizationDTO(BaseModel):
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    form_id: Optional[str] = Field(alias="formId", default=None)
    ip_address: Optional[str] = Field(alias="ipAddress", default=None, hidden=True)
    key: str = Field(alias="key")
    policy_link: Optional[str] = Field(alias="policyLink")
    external_id: Optional[str] = Field(alias="externalId")
    identity_key: str = Field(alias="identityKey")
    identity_value: str = Field(alias="identityValue")
    tags: List[str] = Field(alias="tags", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class AuthorizationSync(GenericSyncResource):
    data: AuthorizationAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "authorizations", header_builder, renew_token, AuthorizationAPIDTO.parse_obj(data))


class AuthorizationAsync(GenericAsyncResource):
    data: AuthorizationAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "authorizations", header_builder, renew_token, AuthorizationAPIDTO.parse_obj(data))


class AuthorizationsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=AuthorizationSync,
                         retrieve_data_model=AuthorizationAPIDTO,
                         create_data_model=CreateAuthorizationDTO,
                         update_data_model=None,
                         resource="authorizations")


class AuthorizationsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=AuthorizationAsync,
                         retrieve_data_model=AuthorizationAPIDTO,
                         create_data_model=CreateAuthorizationDTO,
                         update_data_model=None,
                         resource="authorizations")



from pydantic import BaseModel, Field
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from typing import Optional, Dict


class OTPTemplateAPIDTO(BaseModel):
    id: str = Field(alias="id")
    channel: str = Field(alias="channel")
    label: str = Field(alias="label")
    lang: Optional[str] = Field(alias="lang")
    sender: Optional[str] = Field(alias="sender")
    subject: Optional[str] = Field(alias="subject")
    title: Optional[str] = Field(alias="title")
    header_logo: Optional[str] = Field(alias="headerLogo")
    header_logo_height: Optional[str] = Field(alias="headerLogoHeight", default="40px")
    pre_otp_blurb: Optional[str] = Field(alias="preOtpBlurb")
    pos_otp_blurb: Optional[str] = Field(alias="posOtpBlurb")
    did_not_ask_message: Optional[str] = Field(alias="didNotAskMessage")
    footer_blurb: Optional[str] = Field(alias="footerBlurb")
    footer_link: Optional[str] = Field(alias="footerLink")
    footer_link_label: Optional[str] = Field(alias="footerLinkLabel")
    footer_slogan: Optional[str] = Field(alias="footerSlogan")
    footer_logo: Optional[str] = Field(alias="footerLogo")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateOTPTemplateDTO(BaseModel):
    channel: str = Field(alias="channel")
    label: str = Field(alias="label")
    lang: Optional[str] = Field(alias="lang")
    sender: Optional[str] = Field(alias="sender")
    subject: Optional[str] = Field(alias="subject")
    title: Optional[str] = Field(alias="title")
    header_logo: Optional[str] = Field(alias="headerLogo")
    header_logo_height: Optional[str] = Field(alias="headerLogoHeight", default="40px")
    pre_otp_blurb: Optional[str] = Field(alias="preOtpBlurb")
    pos_otp_blurb: Optional[str] = Field(alias="posOtpBlurb")
    did_not_ask_message: Optional[str] = Field(alias="didNotAskMessage")
    footer_blurb: Optional[str] = Field(alias="footerBlurb")
    footer_link: Optional[str] = Field(alias="footerLink")
    footer_link_label: Optional[str] = Field(alias="footerLinkLabel")
    footer_slogan: Optional[str] = Field(alias="footerSlogan")
    footer_logo: Optional[str] = Field(alias="footerLogo")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class OTPTemplatesSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "otp-templates",header_builder, renew_token, OTPTemplateAPIDTO.parse_obj(data))


class OTPTemplatesAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "otp-templates",header_builder, renew_token, OTPTemplateAPIDTO.parse_obj(data))


class OTPTemplatesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=OTPTemplatesSync,
                         retrieve_data_model=OTPTemplateAPIDTO,
                         create_data_model=CreateOTPTemplateDTO,
                         update_data_model=CreateOTPTemplateDTO,
                         resource="otp-templates")


class OTPTemplatesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=OTPTemplatesAsync,
                         retrieve_data_model=OTPTemplateAPIDTO,
                         create_data_model=CreateOTPTemplateDTO,
                         update_data_model=CreateOTPTemplateDTO,
                         resource="otp-templates")



from pydantic import BaseModel, Field
from typing import List, Optional
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.cms.model.generics import GenericSyncModule, GenericAsyncModule
from altscore.cms.model.common import Money, Schedule, Terms
import datetime as dt
from dateutil.parser import parse as date_parser


class Client(BaseModel):
    id: str = Field(alias="clientId")
    partner_id: str = Field(alias="partnerId")
    external_id: str = Field(alias="externalId")
    email: str = Field(alias="email")
    legal_name: str = Field(alias="legalName")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DPAFlowAPIDTO(BaseModel):
    id: str = Field(alias="flowId")
    tenant: str = Field(alias="tenant")
    reference_id: str = Field(alias="referenceId")
    status: str = Field(alias="status")
    client: Client = Field(alias="client")
    schedule: List[Schedule] = Field(alias="schedule")
    terms: Terms = Field(alias="terms")
    created_at: str = Field(alias="createdAt")
    updated_at: str = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class CreateDPAFlowDTO(BaseModel):
    amount: Money = Field(alias="amount")
    disbursement_date: str = Field(alias="disbursementDate")
    client_id: str = Field(alias="clientId", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class InvoiceInstallment(BaseModel):
    due_date: str = Field(alias="dueDate")
    number: int = Field(alias="number")
    interest: Money = Field(alias="interest")
    amount: Money = Field(alias="amount")
    tax: Money = Field(alias="tax")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Invoice(BaseModel):
    amount: Money = Field(alias="amount")
    invoice_date: str = Field(alias="invoiceDate")
    installments: List[InvoiceInstallment] = Field(alias="installments")
    notes: str = Field(alias="notes")
    reference_id: str = Field(alias="referenceId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class ApproveDPAFlowDTO(BaseModel):
    amount: Money = Field(alias="amount")
    disbursement_date: str = Field(alias="disbursementDate")
    client_id: Optional[str] = Field(alias="clientId", default=None)
    external_id: Optional[str] = Field(alias="externalId", default=None)
    reference_id: Optional[str] = Field(alias="referenceId", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DPABase:

    @staticmethod
    def _approval(flow_id: str):
        return f"/v1/dpas/{flow_id}/approval"

    @staticmethod
    def _cancellation(flow_id: str):
        return f"/v1/dpas/{flow_id}/cancellation"

    @staticmethod
    def _invoice(flow_id: str):
        return f"/v1/dpas/{flow_id}/invoices"


class DPAFlowAsync(DPABase):
    data: DPAFlowAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DPAFlowAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    @retry_on_401_async
    async def approve(self, approve_data: dict):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.put(
                self._approval(self.data.id),
                headers=self._header_builder(),
                json=ApproveDPAFlowDTO.parse_obj(approve_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            self.data = DPAFlowAPIDTO.parse_obj(response.json())

    @retry_on_401_async
    async def cancel(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                self._cancellation(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
            self.data = DPAFlowAPIDTO.parse_obj(response.json())

    @retry_on_401_async
    async def submit_invoice(self, invoice_data: dict):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                self._invoice(self.data.id),
                json=Invoice.parse_obj(invoice_data).dict(by_alias=True),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
            return response.json()

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class DPAFlowSync(DPABase):
    data: DPAFlowAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DPAFlowAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data: DPAFlowAPIDTO = data

    @retry_on_401
    def approve(self, approve_data: dict):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.put(
                self._approval(self.data.id),
                headers=self._header_builder(),
                json=ApproveDPAFlowDTO.parse_obj(approve_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            self.data = DPAFlowAPIDTO.parse_obj(response.json())

    @retry_on_401
    def cancel(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                self._cancellation(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
            self.data = DPAFlowAPIDTO.parse_obj(response.json())

    @retry_on_401
    def submit_invoice(self, invoice_data: dict):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                self._invoice(self.data.id),
                headers=self._header_builder(),
                json=Invoice.parse_obj(invoice_data).dict(by_alias=True),
                timeout=300
            )
            raise_for_status_improved(response)
            return response.json()

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class DPAFlowsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            async_resource=DPAFlowAsync,
            retrieve_data_model=DPAFlowAPIDTO,
            create_data_model=CreateDPAFlowDTO,
            update_data_model=None,
            resource="dpas"
        )

    @retry_on_401_async
    async def create(self, new_entity_data: dict):
        disbursement_date_str = \
            new_entity_data.get("disbursementDate") or new_entity_data.get("disbursement_date")
        if disbursement_date_str is None:
            disbursement_date = dt.date.today().strftime("%Y-%m-%d")
        else:
            disbursement_date = date_parser(disbursement_date_str).strftime("%Y-%m-%d")
        new_entity_data["disbursementDate"] = disbursement_date

        async with httpx.AsyncClient(base_url=self.altscore_client._cms_base_url) as client:
            response = await client.post(
                f"/{self.resource_version}/{self.resource}",
                headers=self.build_headers(),
                json=self.create_data_model.parse_obj(new_entity_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            return self.retrieve_data_model.parse_obj(response.json()).id


class DPAFlowsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            sync_resource=DPAFlowSync,
            retrieve_data_model=DPAFlowAPIDTO,
            create_data_model=CreateDPAFlowDTO,
            update_data_model=None,
            resource="dpas"
        )

    @retry_on_401
    def create(self, new_entity_data: dict):
        disbursement_date_str = \
            new_entity_data.get("disbursementDate") or new_entity_data.get("disbursement_date")
        if disbursement_date_str is None:
            disbursement_date = dt.date.today().strftime("%Y-%m-%d")
        else:
            disbursement_date = date_parser(disbursement_date_str).strftime("%Y-%m-%d")
        new_entity_data["disbursementDate"] = disbursement_date

        with httpx.Client(base_url=self.altscore_client._cms_base_url) as client:
            response = client.post(
                f"/{self.resource_version}/{self.resource}",
                headers=self.build_headers(),
                json=self.create_data_model.parse_obj(new_entity_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            return self.retrieve_data_model.parse_obj(response.json()).id



from pydantic import BaseModel, Field


class Money(BaseModel):
    amount: str
    currency: str


class ScheduleOriginalAmounts(BaseModel):
    fees: Money = Field(alias="fees")
    interest: Money = Field(alias="interest")
    principal: Money = Field(alias="principal")
    taxes: Money = Field(alias="taxes")
    total: Money = Field(alias="total")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Schedule(BaseModel):
    due_date: str = Field(alias="dueDate")
    number: int = Field(alias="number")
    original_amounts: ScheduleOriginalAmounts = Field(alias="originalAmounts")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class InterestRate(BaseModel):
    period: int = Field(alias="period")
    rate: str = Field(alias="rate")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Terms(BaseModel):
    amortization_type: str = Field(alias="amortizationType")
    disbursement_date: str = Field(alias="disbursementDate")
    installments: int = Field(alias="installments")
    interest_rate: InterestRate = Field(alias="interestRate")
    interest_tax: int = Field(alias="interestTax")
    principal: Money = Field(alias="principal")
    repayEvery: int = Field(alias="repayEvery")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True
